/*
 *    **************************************************
 *    *                                                *
 *    *                  BIGPRINT.C                    *
 *    *                                                *
 *    *          Extended Precision Calculator         *
 *    *                                                *
 *    *               Printing Routines                *
 *    *                                                *
 *    *              Version 2.11 03-27-88             *
 *    *                                                *
 *    *              Judson D. McClendon               *
 *    *              844 Sun Valley Road               *
 *    *              Birmingham, AL 35215              *
 *    *                 205-853-8440                   *
 *    *            Compuserve [74415,1003]             *
 *    *                                                *
 *    **************************************************
 */



/*
 *    **************************************************
 *    *                                                *
 *    *                   Includes                     *
 *    *                                                *
 *    **************************************************
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bigcalc.h"
#include "biggvar.h"





/*
 *    **************************************************
 *    *                                                *
 *    *            Source Local Variables              *
 *    *                                                *
 *    **************************************************
 */

static int prevchr;  /* Previous character printed */





/*
 *    **************************************************
 *    *                                                *
 *    *               Printing routines                *
 *    *                                                *
 *    **************************************************
 */



/*
 *    **************************************************
 *    *                                                *
 *    *     Print Registers [lo to hi] on Printer      *
 *    *                                                *
 *    **************************************************
 */

void PrintReg(int lo, int hi)

{

   static int r;

   Message("Printing Registers...");

   PString("\r\n***  Extended Precision Registers  (");
   PChar(PRINTDECIMAL);
   PString(" is decimal point)  ***\r\n\n");

   for (r = lo; r <= hi; r++) {
      PChar(r + '0');
      PrintNumber(&reg[r]);
      NewLine();
      }

   NewLine();

}




/*
 *    **************************************************
 *    *                                                *
 *    *       Print Stack [lo to hi] on Printer        *
 *    *                                                *
 *    **************************************************
 */

void PrintStack(int lo, int hi)

{

   static int s;

   Message("Printing Stack...");

   PString("\r\n***  Extended Precision Stack  (");
   PChar(PRINTDECIMAL);
   PString(" is decimal point)  ***\r\n\n");

   for (s = hi; s >= lo; s--) {
      PChar(stackname[s]);
      PrintNumber(&stack[s]);
      NewLine();
      }

   NewLine();

}




/*
 *    **************************************************
 *    *                                                *
 *    *   Display Memory Register in Full on Screen    *
 *    *                                                *
 *    **************************************************
 */

void DisplayReg(int r)

{

   printtoscreen = TRUE;      /* Temporarily print to screen */

   CurPos(1, 1);
   EraEop();
   menucleared = TRUE;

   PChar(r + '0');
   PrintNumber(&reg[r]);

   MessageWait("");

   WorkScreen();

   printtoscreen = FALSE;     /* Reset print to screen */

}




/*
 *    **************************************************
 *    *                                                *
 *    *    Display Stack Register in Full on Screen    *
 *    *                                                *
 *    **************************************************
 */

void DisplayStack(int s)

{

   printtoscreen = TRUE;      /* Temporarily print to screen */

   CurPos(1, 1);
   EraEop();
   menucleared = TRUE;

   PChar(stackname[s]);
   PrintNumber(&stack[s]);

   MessageWait("");

   WorkScreen();

   printtoscreen = FALSE;     /* Reset print to screen */

}




/*
 *    **************************************************
 *    *                                                *
 *    *       Print Number [*number] on Printer        *
 *    *                                                *
 *    **************************************************
 */

void PrintNumber(NORMTYPE *nbr)

{

   static long exponent;
   static int i, digits, chrcount;

   prevchr = '0';

   PString(": ");
   ppos = SIGNPRTCOL;

   chrcount = 0;
   digits = nbr->digits;
   exponent = nbr->exp;

   if (nbr->digits == 0) {
      PString(" 0\r\n\n");
      return;
      }


   if ( (scinotation)
         ||
        (exponent < minfloatprn)
         ||
        (exponent > maxfloatprn) )

      {                             /* Scientific Notation */

      for (i = 1; i < groupsize; i++)
         PCharWrap(' ');

      PSign(nbr->sign);                   /* Print sign */

      PCharWrap(nbr->man[0]+ '0');        /* First digit and decimal point */
      PCharWrap(PRINTDECIMAL);

      if (digits <= 1)                    /* If no more digits, */
         PCharWrap('0');                  /*  print one zero    */
      else {
         for (i = 1; i < digits; i++) {   /*  print remaining digits */
            PCharWrap(nbr->man[i]+ '0');
            if (! (i % groupsize))
               PCharWrap(' ');
            }
         }
      PExponent(nbr->exp - 1L);           /* Print exponent */

      }                             /* Scientific Notation end */

   else


      {                             /* Floating Decimal */

      if (exponent <= 0L)

         {                                /* Number < 1 */

         for (i = 1; i < groupsize; i++)
            PCharWrap(' ');

         PSign(nbr->sign);                   /* Print sign */
         PCharWrap('0');                     /* Zero left of decimal looks good */
         PCharWrap(PRINTDECIMAL);            /* Decimal point */

         for (i = 0; i > (int)exponent; i--) {  /* Zeros to left of number */
            PCharWrap('0');
            if (! (++chrcount % groupsize))
               PCharWrap(' ');
            }

         for (i = 0; i < digits; i++) {      /* Print mantissa digits */
            PCharWrap(nbr->man[i] + '0');
            if (! (++chrcount % groupsize))
               PCharWrap(' ');
            }
         }                                /* Number < 1 */

      else


         {                                /* Number >= 1 */

                                             /* Calculate digit allignment   */
         i = (int)((3000000L - exponent) % (long)groupsize);
         while (i-- > 0)                     /*  here 1 <= exponent <= maxfloatprn */
            PCharWrap(' ');

         PSign(nbr->sign);                   /* Print sign */

         chrcount = (int)exponent;           /* Distance from decimal point to    */
                                             /*  print groups of groupsize digits */

         for (i = 0; i < (int)exponent; i++) {  /* Digits to left of decimal */
            if (i < digits)
               PCharWrap(nbr->man[i] + '0'); /* Mantissa digits while they last */
            else
               PCharWrap('0');               /* Zeros if mantissa exhausted */

            if (! (--chrcount % groupsize))  /* Space between groups */
               if (chrcount)                 /*  of groupsize digits */
                  PCharWrap(' ');
            }

         chrcount = 0;                       /* Distance from decimal */

         if (i < digits) {                   /* If digits to right of decimal */
            PCharWrap(PRINTDECIMAL);         /*  print decimal point          */

            for (i = (int)exponent; i < digits; i++) {   /* digits to right of decimal */
               PCharWrap(nbr->man[i] + '0');
               if (! (++chrcount % groupsize))
                  PCharWrap(' ');
               }
            }

         }                                /* Number >= 1 */

      }                             /* Floating Decimal */

   NewLine();

}





/*
 *    **************************************************
 *    *                                                *
 *    *         Print Char & Wrap Print Line           *
 *    *                                                *
 *    **************************************************
 */

void PCharWrap(int chr)

{

   static int col;

   if ( (ppos > MAXPRTCOL)
         &&
        (chr != ' ')
         &&
        ( (prevchr == ' ')
           ||
          (prevchr == PRINTDECIMAL) ) ) {
      PString("\r\n");
      for (col = 1; col < MINPRTCOL; col++)
         PChar(' ');
      ppos = MINPRTCOL;
      }

   PChar(chr);
   ppos++;
   prevchr = chr;

}




/*
 *    **************************************************
 *    *                                                *
 *    *              Print Sign Character              *
 *    *                                                *
 *    **************************************************
 */

void PSign(int sign)

{

   if (sign == '-')
      PCharWrap('-');
   else
      PCharWrap(' ');

}




/*
 *    **************************************************
 *    *                                                *
 *    *            Print Exponent on Printer           *
 *    *                                                *
 *    **************************************************
 */

void PExponent(long exponent)

{

   if (prevchr != ' ')
      PChar(' ');
   PChar('e');
   PInteger(exponent);

}





/*
 *    **************************************************
 *    *                                                *
 *    *               Writing routines                 *
 *    *                                                *
 *    **************************************************
 */



/*
 *    **************************************************
 *    *                                                *
 *    *          Write Registers [lo to hi]            *
 *    *                                                *
 *    **************************************************
 */

void WriteReg(int lo, int hi)

{

   static int r;

   for (r = lo; r <= hi; r++) {
      CurPos(r + 4, SIGNDISPCOL);
      WriteNumber(&reg[r]);
      }

}




/*
 *    **************************************************
 *    *                                                *
 *    *            Write Stack [lo to hi]              *
 *    *                                                *
 *    **************************************************
 */

void WriteStack(int lo, int hi)

{

   static int s;

   for (s = hi; s >= lo; s--) {
      CurPos(XSIGNROW - s, SIGNDISPCOL);
      WriteNumber(&stack[s]);
      }

}




/*
 *    **************************************************
 *    *                                                *
 *    *        Write Number [*number] on Screen        *
 *    *                                                *
 *    **************************************************
 */

void WriteNumber(NORMTYPE *nbr)

{
   static long exponent;
   static int i, digits;

   EraEol();

   if (nbr->digits == 0) {
      WString(" 0");
      return;
      }

   if (nbr->sign == '-')
      WChar('-');
   else
      WChar(' ');
   ppos = MINDISPCOL;

   digits = nbr->digits;
   exponent = nbr->exp;

   if ( (scinotation)
         ||
        (exponent < MINFLOATDSP)
         ||
        (exponent > MAXFLOATDSP) )

      {                             /* Scientific Notation */

      WChar(nbr->man[0]+ '0');         /* First digit and decimal point */
      WChar(DISPDECIMAL);
      ppos += 2;

      if (digits <= 1) {                  /* If no more digits, */
         WChar('0');                      /*  print one zero    */
         ppos++;
         }
      else
         for (i = 1; i < digits; i++) {   /*  print remaining digits */
            WChar(nbr->man[i]+ '0');
            if (++ppos > (MAXDISPCOL - 1))
               break;
            }
      WExponent(nbr->exp - 1L);           /* Write exponent */

      }                             /* Scientific Notation end */

   else

      {                             /* Floating Decimal */
      if (exponent <= 0L)

         {                             /* Number < 1 */

         WChar(DISPDECIMAL);                 /* Decimal point */
         ppos++;

         for (i = 0; i > (int)exponent; i--) {  /* Zeros to left of number */
            WChar('0');
            ppos++;
            }

         for (i = 0; i < digits; i++) {      /* Write number digits */
            WChar(nbr->man[i] + '0');
            if (++ppos > MAXDISPCOL)         /* Until end of digits or line */
               break;
            }
         }                             /* Number < 1 end */

      else

         {                             /* Number >= 1 */

         for (i = 0; i < (int)exponent; i++) {  /* Digits to left of decimal */
            if (i < digits)
               WChar(nbr->man[i] + '0');  /* Mantissa digits while they last */
            else
               WChar('0');                /* Zeros if mantissa exhausted */
            ppos++;
            }

         if (i < digits) {                /* Write digits to right of decimal */
            WChar(DISPDECIMAL);
            ppos++;
            for (i = (int)exponent; i < digits; i++) {
               WChar(nbr->man[i] + '0');
               if (++ppos > MAXDISPCOL)
                  break;
               }
            }
         }                             /* Number >= 1 end */

      }                             /* Floating Decimal end */

}




/*
 *    **************************************************
 *    *                                                *
 *    *            Write Exponent on Screen            *
 *    *                                                *
 *    **************************************************
 */

void WExponent(long exponent)

{

   WString(" e");
   WInteger(exponent);

}




/*
 *    **************************************************
 *    *                                                *
 *    *             Get Prompted Character             *
 *    *                                                *
 *    **************************************************
 */

int GetPrompt(void)

{

   if (menucleared) {            /* Restore menu if cleared */
      OnScreenMenu();
      menucleared = FALSE;
      }

   if (charpresent) {            /* Character passed from routine */
      charpresent = FALSE;
      return(chr);
      }
   else {                        /* Get character */
      Message("(Esc to Exit)");
      return(GetChar());
      }

}




/*
 *    **************************************************
 *    *                                                *
 *    *            Write Overflow Message              *
 *    *                                                *
 *    **************************************************
 */

void Overflow(void)

{

   Beep();
   MessageWait("** Overflow **");

}




/*
 *    **************************************************
 *    *                                                *
 *    *          Write Divide By Zero Message          *
 *    *                                                *
 *    **************************************************
 */

void DivideByZero(void)

{

   Beep();
   MessageWait("** Divide by zero **");

}




/*
 *    **************************************************
 *    *                                                *
 *    *        Write Negative Argument Message         *
 *    *                                                *
 *    **************************************************
 */

void NegativeArgument(void)

{

   Beep();
   MessageWait("** Negative Argument **");

}




/*
 *    **************************************************
 *    *                                                *
 *    *       Write Argument not Integer Message       *
 *    *                                                *
 *    **************************************************
 */

void ArgumentNotInteger(void)

{

   Beep();
   MessageWait("** Argument not Integer **");

}




/*
 *    **************************************************
 *    *                                                *
 *    *                Screen Routines                 *
 *    *                                                *
 *    **************************************************
 */




/*
 *    **************************************************
 *    *                                                *
 *    *                On Screen Menu                  *
 *    *                                                *
 *    **************************************************
 */

void OnScreenMenu(void)

{

   CurPos(19, 1);
   EraEop();
   WriteAt(20, 1, " 0-9.E  Number   + >Add        X >Xchg X R "
                  "  V >View full    F1 >Help  F2 >ABS");
   WriteAt(21, 1, "     S  ChgSign  - >Sub   Lft/Rt >Xchg X Y "
                  "  F >Float/Sci    F3 >     F4 >X");
   WriteAt(22, 1, "BakSpc  Clear X  * >Mul   Up/Dwn >Roll UpDn"
                  "  G >Group 3/5    F5 >1/X   F6 >X!");
   WriteAt(23, 1, " Enter >Enter    / >Div     PgUp >Store    "
                  "  P >Print        F7 >INT   F8 >FRAC");
   WriteAt(24, 1, "     L >Last X   C >Clear   PgDn >Recall   "
                  "  D >Disk/Printr  F9 >    F10 >e");

}




/*
 *    **************************************************
 *    *                                                *
 *    *                Initial Screen                  *
 *    *                                                *
 *    **************************************************
 */

void InitialScreen(void)

{

   HelpScreen1();
   MessageWait("");

   WorkScreen();

}




/*
 *    **************************************************
 *    *                                                *
 *    *                  Help Screen                   *
 *    *                                                *
 *    **************************************************
 */

void HelpScreen(void)

{

   static int screen;

   menucleared = TRUE;

   screen = 2;

   do {
      switch (screen) {
         case (1):
            HelpScreen1();
            break;
         case (2):
            HelpScreen2();
            break;
         case (3):
            HelpScreen3();
            break;
         }  /* switch */

      Message("(Press +=Next page, -=Prev page, Esc=Exit to BIGCALC)");

      while (TRUE) {
         chr = GetChar();
         if (chr == '+') {
            if (++screen > 3)
               screen = 1;
            break;
            }
         else if (chr == '-') {
            if (--screen < 1)
               screen = 3;
            break;
            }
         else if (chr == ESCAPE)
            break;
         else
            Beep();
         }  /* while (TRUE) */

      } while (chr != ESCAPE);

   WorkScreen();

}




/*
 *    **************************************************
 *    *                                                *
 *    *             Screen Header Line 1               *
 *    *                                                *
 *    **************************************************
 */

void Heading1(void)

{

   ScrClr();
   WriteCenter(1, TITLE);

}




/*
 *    **************************************************
 *    *                                                *
 *    *             Screen Header Line 2               *
 *    *                                                *
 *    **************************************************
 */

void Heading2(void)

{

   CurPos(2, 1);
   EraEol();
   WriteAt(2, 8, "Precision is ");
   WInteger((long) normprec);
   WString(" digits,  Digit grouping is ");
   WInteger((long) groupsize);
   WString(",  Print to ");
   WString(printid);

}




/*
 *    **************************************************
 *    *                                                *
 *    *             Display Help Screen 1              *
 *    *                                                *
 *    **************************************************
 */

void HelpScreen1(void)

{

   Heading1();

   WriteAt( 3,  1, "BIGCALC emulates a H-P calculator with stack and ten"
                   " memory registers with");
   WriteAt( 4,  1, "maximum precision settable from ");
   WInteger((long)MINPREC);
   WString(" to ");
   WInteger((long)NORMPREC);
   WString(" digits. Exponents can range ");
   WInteger(MAXEXP);
   WString(".");
   WriteAt( 5,  1, "Execute BIGCALC like this:");

   WriteAt( 7,  3, "BIGCALC precision   (precision is maximum number of "
                   "digits, default is ");
   WInteger((long)DEFAULTPREC);
   WString(")");

   WriteAt( 9,  1, "BIGCALC has add, subtract, multiply, divide, square,"
                   " square root, factorial,");
   WriteAt(10,  1, "reciprocal, change sign, pi, e, print, store, recall"
                   ", and much more. BIGCALC is");
   WriteAt(11,  1, "very fast, but operations on large numbers may take "
                   "several seconds. You can");
   WriteAt(12,  1, "abort long calculations by pressing the Escape key. "
                   "Call up on screen help by");
   WriteAt(13,  1, "pressing the F1 key. You can send print output to di"
                   "sk for other uses.");

   WriteAt(15,  1, "You can specify either floating decimal or scientifi"
                   "c notation display. Large");
   WriteAt(16,  1, "numbers display in scientific notation to about 70 d"
                   "igits all the time, but the");
   WriteAt(17,  1, "view and print commands show full precision in 3 or "
                   "5 digit groups.");

   WriteAt(19,  1, " Judson D. McClendon       If you like this program "
                   "consider sending a small");
   WriteAt(20,  1, " 844 Sun Valley Road       contribution to the autho"
                   "r. Comments are welcome.");
   WriteAt(21,  1, " Birmingham, AL 35215      BIGCALC is about 5800 lin"
                   "es of C.");
   WriteAt(22,  1, "     205-853-8440          There is no warranty of a"
                   "ny kind. The author assumes");
   WriteAt(23,  1, "Compuserve [74415,1003]    no responsibility for the"
                   " use of this program.");

}




/*
 *    **************************************************
 *    *                                                *
 *    *             Display Help Screen 2              *
 *    *                                                *
 *    **************************************************
 */

void HelpScreen2(void)

{

   Heading1();

   WriteAt( 3,  1, "   KEY  FUNCTION           KEY  FUNCTION           "
                   "  KEY  FUNCTION");
   WriteAt( 5,  1, " 0-9.E  Enter Number         V >View full number   "
                   "   F1 >Help");
   WriteAt( 6,  1, "     S  change Sign          F >toggle Float/Sci   "
                   "   F3 >Square root (x)");
   WriteAt( 7,  1, "BakSpc  clear X/backspace    G >toggle Group 3/5   "
                   "   F5 >Reciprocal (1  X)");
   WriteAt( 8,  1, " Enter >Enter                P >Print              "
                   "   F7 >Integer part (XXX.)");
   WriteAt( 9,  1, "     + >Add (Y + X)          D >toggle Disk/Printer"
                   "   F9 >Recall  const to X");
   WriteAt(10,  1, "     - >Sub (Y - X)       PgUp >Store X in Reg     "
                   "   F2 >Absolute value X");
   WriteAt(11,  1, "     * >Mul (Y x X)       PgDn >Recall Reg to X    "
                   "   F4 >Square (X)");
   WriteAt(12,  1, "     / >Div (Y  X)          X >eXchange X & Reg   "
                   "   F6 >Factorial (X!)");
   WriteAt(13,  1, "     C >Clear          Left/Rt >Exchange X & Y     "
                   "   F8 >Fraction part (.XXX)");
   WriteAt(14,  1, "     L >Last X         Up/Down >Roll Stack Up/Down "
                   "  F10 >Recall e const to X");

   WriteAt(16,  5, "To enter a number just begin typeing it in. Use Bac"
                   "kspace to back up.");
   WriteAt(17,  5, "To enter an exponent press E then enter the exponen"
                   "t.");
   WriteAt(18,  5, "Press S while entering mantissa or exponent to chan"
                   "ge respective sign.");
   WriteAt(19,  5, "If you start a number with E a mantissa of 1 is ass"
                   "umed.");
   WriteAt(20,  5, "Functions marked '>' complete a number being entere"
                   "d before acting.");
   WriteAt(21,  5, "View displays any register to full precision.");
   WriteAt(22,  5, "Clear, Print, View, Store, Recall & eXchange ask fo"
                   "r affected register.");
   WriteAt(23,  5, "BackSpace backs up while entering number, otherwise"
                   " acts as Clear X.");

}




/*
 *    **************************************************
 *    *                                                *
 *    *             Display Help Screen 3              *
 *    *                                                *
 *    **************************************************
 */

void HelpScreen3(void)

{

   Heading1();

   WriteAt( 3,  1, "This program emulates a Hewlett-Packard calculator with"
                   " four register stack");
   WriteAt( 4,  1, "and ten memory registers.  The stack consists of the X,"
                   " Y, Z & T registers.");
   WriteAt( 5,  1, "The memory registers are 0-9.");

   WriteAt( 7,  1, "The X register is 'where the action is'.  When you ente"
                   "r a number it is entered");
   WriteAt( 8,  1, "into the X register.  If X already contains a number, t"
                   "hat number is 'pushed'");
   WriteAt( 9,  1, "up to Y, Y is pushed to Z, Z to T, and the contents of "
                   "T is lost.  The 'Enter'");
   WriteAt(10,  1, "key pushes the stack and duplicates X into Y. A number "
                   "entered into X after");
   WriteAt(11,  1, "Enter or ClearX DOES NOT push the stack. When you use a"
                   " function that acts on");
   WriteAt(12,  1, "two numbers (like + or -) it always acts on X & Y (exce"
                   "pt for Store, Recall &");
   WriteAt(13,  1, "eXchange which act on X and a memory register).  The re"
                   "sult is put in X, and");
   WriteAt(14,  1, "the stack is 'Dropped'.  That is, Z copies to Y and T d"
                   "uplicates into Z.");

   WriteAt(16,  1, "Numbers are displayed in 'Floating Decimal' or 'Scienti"
                   "fic Notation' format.");
   WriteAt(17,  1, "Floating Decimal looks like this:    1.2     123.45    "
                   " -12300000000  0.000002");
   WriteAt(18,  1, "Scientific Notation looks like this: 1.2 e0  1.2345 e2 "
                   " -1.23 e10     2.0 e-6");
   WriteAt(19,  1, "Very long numbers take longer to compute, of course. Yo"
                   "u can abort a long");
   WriteAt(20,  1, "calculation by pressing the Escape key.");

   WriteAt(22,  1, "If a number is too large or small for the screen, Scient"
                   "ific Notation is used.");
   WriteAt(23,  1, "Calculations are performed to twice the set precision th"
                   "en truncated.");

}




/*
 *    **************************************************
 *    *                                                *
 *    *              Display Work Screen               *
 *    *                                                *
 *    **************************************************
 */

void WorkScreen(void)

{

   static int r, s;

   Heading1();
   Heading2();

   WriteAt(3, 1, "=========================="
                 "  R E G I S T E R S  "
                 "==========================");
   for (r = 0; r <= 9; r++) {
      CurPos(4 + r, 1);
      WChar(r + '0');
      WChar(':');
      WriteReg(r, r);
   }

   WriteAt(14, 1, "=============================="
                  "  S T A C K  "
                  "==============================");
   for (s = 3; s >= 0; s--) {
      CurPos(XSIGNROW - s, 1);
      WChar(stackname[s]);
      WChar(':');
      WriteStack(s, s);
   }

}
